/*
 *  ALSA PCM Tester
 *
 *  Copyright (c) 2013 Eldad Zack <eldad@fogrefinery.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <alsa/asoundlib.h>
#include <pthread.h>
#include <sys/time.h>

#include "alsa_pcm_tests.h"
#define lengthof(x) ((size_t)(sizeof(x) / sizeof(__typeof__(*x))))

#define ERR(err, arg...) do { if (err < 0) { printf(arg); } } while (0)
#define ERR_ON(err, msg, arg...) do { if (err < 0) { fprintf(stderr, "%s:%d: error %d on " msg "\n" , __func__, __LINE__, err, arg); } } while (0)

#define STATE_NOOP	0
#define STATE_START	1
#define STATE_STOP	2
#define STATE_STOPPED	3

static bool io_err[2];
static pthread_t pth[2];
static bool th_state[2];
static long int frames[2];

struct pcm_handle_info {
	int idx;
	snd_pcm_t *pcm_handle;
	snd_pcm_stream_t stream;
	char *name;
	int channels;
	void* (*loop_op)();
};

char *hw_name;
char *shell_cmd;
unsigned int rate;
unsigned int waitsleep_ms;

void *playback_thread();
void *capture_thread();

static struct pcm_handle_info phis[] = {
		{ 0, NULL, SND_PCM_STREAM_PLAYBACK, "playback", 0, &playback_thread}, 
		{ 1, NULL, SND_PCM_STREAM_CAPTURE, "capture", 0, &capture_thread}
	};

static int run_user_cmd()
{
	int err;

	if (!shell_cmd)
		return 0;

	printf(" * Executing %s\n", shell_cmd);
	err = system(shell_cmd);
	ERR_ON(err, "user command %s", shell_cmd);

	return err;
}

static int pcm_open(struct pcm_handle_info *phi)
{
	int err;

	err = snd_pcm_open(&phi->pcm_handle, hw_name, phi->stream, 0);
	ERR_ON(err, "stream %d", phi->stream);

	return err;
}

static int pcm_close(struct pcm_handle_info *phi)
{
	int err;

	err = snd_pcm_close(phi->pcm_handle);
	ERR_ON(err, "name %s", phi->name);

	return err;
}

static int pcm_hw_params(struct pcm_handle_info *phi)
{
	int err;
	unsigned int ptmin = 50 * 1000;
	snd_pcm_hw_params_t *hw_params;

	err = snd_pcm_hw_params_malloc(&hw_params);
	ERR_ON(err, "%s", "malloc");
	if (err < 0)
		return err;

	err = snd_pcm_hw_params_any(phi->pcm_handle, hw_params);
	ERR_ON(err, "%s", "any");
	if (err < 0)
		return err;

	err = snd_pcm_hw_params_set_access(phi->pcm_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED);
	ERR_ON(err, "%s", "set_access");
	if (err < 0)
		return err;

	err = snd_pcm_hw_params_set_rate_near(phi->pcm_handle, hw_params, &rate, NULL);
	ERR_ON(err, "%s", "set_rate_near");
	if (err < 0)
		return err;

	err = snd_pcm_hw_params_set_period_time_min(phi->pcm_handle, hw_params, &ptmin, NULL);
	ERR_ON(err, "%s", "set_rate_near");
	if (err < 0)
		return err;

	err = snd_pcm_hw_params(phi->pcm_handle, hw_params);
	ERR_ON(err, "%s", "hw_params");
	if (err < 0)
		return err;

	snd_pcm_hw_params_free(hw_params);
	return 0;
}

static int pcm_prepare(struct pcm_handle_info *phi)
{
	int err;

	err = snd_pcm_prepare(phi->pcm_handle);
	ERR_ON(err, "%s", "prepare");

	return err;
}

void *playback_thread()
{
	int ret;
	snd_pcm_uframes_t buffer_size;
	snd_pcm_uframes_t period_size;

	char *bufp;

	ret = snd_pcm_get_params(phis[0].pcm_handle, &buffer_size, &period_size);
	ERR_ON(ret, "%s", "get_params");

	bufp = malloc(buffer_size);
	if (bufp == NULL) {
		printf("   == playback malloc error\n");
		exit(-1);
	}
	memset(bufp, 0, buffer_size);

	while (th_state[0] == STATE_START) {
		ret = snd_pcm_writei(phis[0].pcm_handle, bufp, rate / 100);
		if (ret < 0)
			io_err[0] = true;
		ERR_ON(ret, "writei %ld", frames[0]);

		frames[0] += ret;

		ret = snd_pcm_wait(phis[0].pcm_handle, -1);
		if (ret == 0) {
			printf("   == Timeout: playback (out: %ld) period %ld buffer %ld\n", frames[0], period_size, buffer_size);
		}
	}

	th_state[0] = STATE_STOPPED;
	return NULL;
}

void *capture_thread()
{
	int ret;
	snd_pcm_uframes_t buffer_size;
	snd_pcm_uframes_t period_size;

	char *bufp;

	ret = snd_pcm_get_params(phis[1].pcm_handle, &buffer_size, &period_size);
	ERR_ON(ret, "%s", "get_params");

	bufp = malloc(buffer_size);
	if (bufp == NULL) {
		printf("   == capture malloc error\n");
		exit(-1);
	}
	memset(bufp, 0, buffer_size);

	while (th_state[1] == STATE_START) {
		ret = snd_pcm_readi(phis[1].pcm_handle, bufp, rate / 100);
		if (ret < 0)
			io_err [1] = true;
		ERR_ON(ret, "readi %ld", frames[1]);

		frames[1] += ret;

		ret = snd_pcm_wait(phis[1].pcm_handle, -1);
		if (ret == 0) {
			printf("   == Timeout: capture (in: %ld) period %ld buffer %ld\n", frames[1], period_size, buffer_size);
		}
	}

	th_state[1] = STATE_STOPPED;
	return NULL;
}

static int pcm_start(struct pcm_handle_info *phi)
{
	frames[phi->idx] = 0;
	th_state[phi->idx] = STATE_START;
	pthread_create(&pth[phi->idx], NULL, phi->loop_op, NULL);

	return 0;
}

static int waitsleep(struct pcm_handle_info *phi)
{
	usleep(waitsleep_ms * 1000);

	return 0;
}

static int pcm_stop(struct pcm_handle_info *phi)
{
	th_state[phi->idx] = STATE_STOP;
	pthread_cancel(pth[phi->idx]);
	printf("  ++ Frames [%s]: %ld\n", phi->name, frames[phi->idx]);

	if (frames[phi->idx] == 0)
		return -EPIPE;
	return 0;
}

static bool valid_perm(unsigned long permn, int opsn)
{
	unsigned long bit;
	int setbits = 0;

	for (bit = 1; bit < (1 << opsn); bit <<= 1) {
		if (permn & bit)
			setbits++;
	}

	return (setbits == (opsn >> 1));
}

struct step_op {
	int (*op)(struct pcm_handle_info *);
	const char *name;
};

static int run_steps(int steps, unsigned int order, struct step_op step_ops[])
{
	int err;
	int i;
	int tgt;
	int step[2] = {0 ,0};
	unsigned long bit = 1;

	io_err[0] = false;
	io_err[1] = false;
	for (i = 0; i < (steps * 2); i++) {
		tgt = (bit & order) == 0 ? 0 : 1;

		printf("  * [%d/%d] %s => %s\n", i + 1, steps * 2, phis[tgt].name, step_ops[step[tgt]].name);
		err = step_ops[step[tgt]].op(&phis[tgt]);
		if (err < 0)
			return err;

		step[tgt]++;
		bit <<= 1;

		if (io_err[0]) {
			printf(" ** IO Error (%s)\n", phis[0].name);
			return -EIO;
		}
		if (io_err[1]) {
			printf(" ** IO Error (%s)\n", phis[1].name);
			return -EIO;
		}
	}

	return 0;
}

static void run_test(int steps, struct step_op step_ops[], const char *test_name, long unsigned int permn, long unsigned int last_permn)
{
	int err;
	unsigned long i;

	printf(" * Test %s, Steps: ", test_name);
	for (i = 0; i < steps; i++) {
		printf(" [%s]", step_ops[i].name);
	}
	printf("\n");

	if (permn < 3)
		permn = 3;
	
	if (last_permn == 0)
		last_permn = 1 << (steps * 2);
	last_permn++;

	for (i = permn; i < last_permn; i++) {
		if (valid_perm(i, (steps * 2))) {
			struct timeval start, stop, delta;
			gettimeofday(&start, NULL);

			printf("++ Test: %s, permutation: %#08lx\n", test_name, i);
			err = run_user_cmd();
			if (err < 0) {
				printf("user command failed on test %s permutation %08lx, exiting.\n", test_name, i);
				exit(-1);
			}

			usleep(50 * 1000);
			err = run_steps(steps, i, step_ops);

			gettimeofday(&stop, NULL);
			timersub(&stop, &start, &delta);
			printf("++ permutation: %#08lx, runtime: %ld.%06ld sec\n", i, delta.tv_sec, delta.tv_usec);

			if (err < 0) {
				printf("!! Test %s permutation %#08lx failed !!\n", test_name, i);
				exit(-1);
			}
		}
	}
}

#define STEP_OP(func) { .op = &func, .name = #func }
#define DECLARE_TEST(name, steps...) void test_ ## name (long unsigned int permn, long unsigned int last_permn) { \
	struct step_op ops[] = { steps }; \
	run_test(lengthof(ops), ops, #name, permn, last_permn); \
}

DECLARE_TEST(open,
	STEP_OP(pcm_open),
	STEP_OP(pcm_close));

DECLARE_TEST(hw_params,
	STEP_OP(pcm_open),
	STEP_OP(pcm_hw_params),
	STEP_OP(pcm_close));

DECLARE_TEST(prepare,
	STEP_OP(pcm_open),
	STEP_OP(pcm_hw_params),
	STEP_OP(pcm_prepare),
	STEP_OP(pcm_close));

DECLARE_TEST(stream,
	STEP_OP(pcm_open),
	STEP_OP(pcm_hw_params),
	STEP_OP(pcm_prepare),
	STEP_OP(pcm_start),
	STEP_OP(waitsleep),
	STEP_OP(pcm_stop),
	STEP_OP(pcm_close));

