/*
 *  ALSA PCM Tester
 *
 *  Copyright (c) 2013 Eldad Zack <eldad@fogrefinery.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "alsa_pcm_tests.h"

#define lengthof(x) ((size_t)(sizeof(x) / sizeof(__typeof__(*x))))

void test_all(unsigned long first, unsigned long last)
{
	test_open(0,0);
	test_hw_params(0,0);
	test_prepare(0,0);
	test_stream(0,0);
}

struct test_desc {
	const char *name;
	const char *desc;
	void (*proc)(unsigned long, unsigned long);
};

struct test_desc test_descs[] = {
	{
		.name = "all",
		.desc = "run all tests",
		.proc = &test_all,
	},
	{
		.name = "open",
		.desc = "open and close device",
		.proc = &test_open,
	},
	{
		.name = "hw_params",
		.desc = "open, set hw parameters and close device",
		.proc = &test_hw_params,
	},
	{
		.name = "prepare",
		.desc = "open, set hw parameters, prepare and close device (no stream)",
		.proc = &test_prepare,
	},
	{
		.name = "stream",
		.desc = "like prepare, but also start a stream",
		.proc = &test_stream,
	},
};

static void print_help()
{
	int i;

	printf("Usage:\n atest [options] <test>\n");
	printf("* --device            ALSA PCM device (e.g., hw:0)\n");
	printf("* --rate              PCM Rate (e.g., 48000)\n");
	printf("  --first             First permutation ID\n");
	printf("  --last              Last permutation ID\n");
	printf("  --waitsleep         Duration of capture/playback waitsleep\n");
	printf("  --cmd               Run this command between permutations\n");
	printf("\n(*) Mandatory arguments\n");

	printf("\n<test> is one of the following:\n");

	for (i = 0; i < lengthof(test_descs); i++)
		printf(" %-10s %s\n", test_descs[i].name, test_descs[i].desc);

	printf("\n");
}

#define OPT_HW		1
#define OPT_RATE	2
#define OPT_FIRST	3
#define OPT_LAST	4
#define OPT_WAITSLEEP	5
#define OPT_SHELL_CMD	6
#define OPT_HELP	7

static unsigned long first = 0;
static unsigned long last = 0;

static void parse_options(int argc, char *argv[])
{
	int opt, i;

	static const struct option longopts[] = {
		{
			.name = "device",
			.has_arg = required_argument,
			.flag = NULL,
			.val = OPT_HW,
		},
		{
			.name = "rate",
			.has_arg = required_argument,
			.flag = NULL,
			.val = OPT_RATE,
		},
		{
			.name = "first",
			.has_arg = required_argument,
			.flag = NULL,
			.val = OPT_FIRST,
		},
		{
			.name = "last",
			.has_arg = required_argument,
			.flag = NULL,
			.val = OPT_LAST,
		},
		{
			.name = "waitsleep",
			.has_arg = required_argument,
			.flag = NULL,
			.val = OPT_WAITSLEEP,
		},
		{
			.name = "cmd",
			.has_arg = required_argument,
			.flag = NULL,
			.val = OPT_SHELL_CMD,
		},
		{
			.name = "help",
			.has_arg = no_argument,
			.flag = NULL,
			.val = OPT_HELP,
		},
		{ 0 }
	};

	opterr = 0;
	while ((opt = getopt_long_only(argc, argv, "", longopts, NULL)) != -1) {
		switch (opt) {
		case OPT_HW:
			hw_name = optarg;
			break;
		case OPT_RATE:
			rate = atoi(optarg);
			break;
		case OPT_FIRST:
			first = strtol(optarg, NULL, 0);
			break;
		case OPT_LAST:
			last = strtol(optarg, NULL, 0);
			break;
		case OPT_WAITSLEEP:
			waitsleep_ms = atoi(optarg);
			break;
		case OPT_SHELL_CMD:
			shell_cmd = optarg;
			break;
		case OPT_HELP:
			print_help();
			exit(0);
		default:
			printf("Unknown option: %s\n", argv[optind - 1]);
			exit(0);
		}
	}

	if (optind >= argc) {
		printf("Please specify test. (use -h for help)\n");
		exit(0);
	}

	if (hw_name == NULL) {
		printf("Please specify device (e.g., --device hw:0)\n");
		exit(0);
	}

	if (rate == 0)
		rate = 96000;
	if (waitsleep_ms == 0)
		waitsleep_ms = 75;

	printf("Testing device %s, rate %d, range %lx:%lx\n", hw_name, rate, first, last);

	for (i = 0; i < lengthof(test_descs); i++) {
		if (strcmp(argv[optind], test_descs[i].name) == 0) {
			optind++;
			break;
		}
	}

	if (optind != argc) {
		printf("Invalid invocation (%d : %d).\n", optind, argc);
		print_help();
		exit(0);
	}

	test_descs[i].proc(first, last);
}

int main(int argc, char *argv[])
{
	printf("atest v%s\n", VERSION);

	parse_options(argc, argv);

	return 0;
}
